create_history = """CREATE TABLE IF NOT EXISTS media (domain TEXT,
                                               url_path TEXT,
                                               referer TEXT,
                                               download_path TEXT,
                                               download_filename TEXT,
                                               original_filename TEXT,
                                               completed INTEGER NOT NULL,
                                               PRIMARY KEY (domain, url_path, original_filename)
                                               );"""

create_fixed_history = """CREATE TABLE IF NOT EXISTS media_copy (domain TEXT,
                                               url_path TEXT,
                                               referer TEXT,
                                               download_path TEXT,
                                               download_filename TEXT,
                                               original_filename TEXT,
                                               completed INTEGER NOT NULL,
                                               PRIMARY KEY (domain, url_path, original_filename)
                                               );"""

create_temp = """CREATE TABLE IF NOT EXISTS temp (downloaded_filename TEXT);"""



from __future__ import annotations

from typing import TYPE_CHECKING

from aiolimiter import AsyncLimiter
from yarl import URL

from cyberdrop_dl.scraper.crawler import Crawler
from cyberdrop_dl.utils.dataclasses.url_objects import ScrapeItem
from cyberdrop_dl.utils.utilities import error_handling_wrapper, get_filename_and_ext

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


class RedGifsCrawler(Crawler):
    def __init__(self, manager: Manager):
        super().__init__(manager, "redgifs", "RedGifs")
        self.redgifs_api = URL("https://api.redgifs.com/")
        self.token = ""
        self.headers = {}
        self.request_limiter = AsyncLimiter(10, 1)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def fetch(self, scrape_item: ScrapeItem) -> None:
        """Determines where to send the scrape item based on the url"""
        task_id = await self.scraping_progress.add_task(scrape_item.url)

        if not self.token:
            await self.manage_token(self.redgifs_api / "v2/auth/temporary")

        if self.token:
            if "users" in scrape_item.url.parts:
                await self.user(scrape_item)
            else:
                await self.post(scrape_item)

        await self.scraping_progress.remove_task(task_id)

    @error_handling_wrapper
    async def user(self, scrape_item: ScrapeItem) -> None:
        """Scrapes a users page"""
        user_id = scrape_item.url.parts[-1].split(".")[0]

        page = 1
        total_pages = 1
        while page <= total_pages:
            async with self.request_limiter:
                JSON_Resp = await self.client.get_json(self.domain, (self.redgifs_api / "v2/users" / user_id / "search").with_query(f"order=new&count=40&page={page}"), headers_inc=self.headers)
            total_pages = JSON_Resp["pages"]
            gifs = JSON_Resp["gifs"]
            for gif in gifs:
                links = gif["urls"]
                date = gif["createDate"]
                title_part = gif["title"] if "title" in gif else f"Loose Files"
                title = await self.create_title(title_part, None, None)

                try:
                    link = URL(links["hd"])
                except (KeyError, TypeError):
                    link = URL(links["sd"])

                filename, ext = await get_filename_and_ext(link.name)
                new_scrape_item = await self.create_scrape_item(scrape_item, link, title, True, date)
                await self.handle_file(link, new_scrape_item, filename, ext)
            page += 1

    @error_handling_wrapper
    async def post(self, scrape_item: ScrapeItem) -> None:
        """Scrapes a post"""
        post_id = scrape_item.url.parts[-1].split(".")[0]

        async with self.request_limiter:
            JSON_Resp = await self.client.get_json(self.domain, self.redgifs_api / "v2/gifs" / post_id, headers_inc=self.headers)

        title_part = JSON_Resp["gif"]["title"] if "title" in JSON_Resp["gif"] else "Loose Files"
        title = await self.create_title(title_part, None, None)
        links = JSON_Resp["gif"]["urls"]
        date = JSON_Resp["gif"]["createDate"]

        if "hd" in links:
            link = URL(links["hd"])
        else:
            link = URL(links["sd"])

        filename, ext = await get_filename_and_ext(link.name)
        new_scrape_item = await self.create_scrape_item(scrape_item, link, title, True, date)
        await self.handle_file(link, new_scrape_item, filename, ext)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    @error_handling_wrapper
    async def manage_token(self, token_url: URL) -> None:
        """Gets/Sets the redgifs token and header"""
        async with self.request_limiter:
            json_obj = await self.client.get_json(self.domain, )
        self.token = json_obj["token"]
        self.headers = {"Authorization": f"Bearer {self.token}"}



import os
from dataclasses import field
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager

if os.getenv("PYCHARM_HOSTED") is not None:
    """This is for testing purposes only"""
    APP_STORAGE = Path("../AppData")
    DOWNLOAD_STORAGE = Path("../Downloads")
else:
    APP_STORAGE = Path("./AppData")
    DOWNLOAD_STORAGE = Path("./Downloads")


class PathManager:
    def __init__(self, manager: 'Manager'):
        self.manager = manager

        self.download_dir: Path = field(init=False)
        self.sorted_dir: Path = field(init=False)
        self.log_dir: Path = field(init=False)

        self.cache_dir: Path = field(init=False)
        self.config_dir: Path = field(init=False)

        self.input_file: Path = field(init=False)
        self.history_db: Path = field(init=False)

        self.main_log: Path = field(init=False)
        self.last_post_log: Path = field(init=False)
        self.unsupported_urls_log: Path = field(init=False)
        self.download_error_log: Path = field(init=False)
        self.scrape_error_log: Path = field(init=False)

    def pre_startup(self) -> None:
        if self.manager.args_manager.appdata_dir:
            global APP_STORAGE
            APP_STORAGE = Path(self.manager.args_manager.appdata_dir) / "AppData"

        self.cache_dir = APP_STORAGE / "Cache"
        self.config_dir = APP_STORAGE / "Configs"

        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.config_dir.mkdir(parents=True, exist_ok=True)

    def startup(self) -> None:
        """Startup process for the Directory Manager"""
        self.download_dir = self.manager.config_manager.settings_data['Files']['download_folder']
        self.sorted_dir = self.manager.config_manager.settings_data['Sorting']['sort_folder']
        self.log_dir = self.manager.config_manager.settings_data['Logs']['log_folder'] if not self.manager.args_manager.log_dir else self.manager.args_manager.log_dir
        self.input_file = self.manager.config_manager.settings_data['Files']['input_file']
        self.history_db = self.cache_dir / "cyberdrop.db"

        self.main_log = self.log_dir / self.manager.config_manager.settings_data['Logs']['main_log_filename']
        self.last_post_log = self.log_dir / self.manager.config_manager.settings_data['Logs']['last_forum_post_filename']
        self.unsupported_urls_log = self.log_dir / self.manager.config_manager.settings_data['Logs']['unsupported_urls_filename']
        self.download_error_log = self.log_dir / self.manager.config_manager.settings_data['Logs']['download_error_urls_filename']
        self.scrape_error_log = self.log_dir / self.manager.config_manager.settings_data['Logs']['scrape_error_urls_filename']

        self.log_dir.mkdir(parents=True, exist_ok=True)
        if not self.input_file.is_file() and not self.manager.args_manager.input_file:
            self.input_file.touch(exist_ok=True)
        self.history_db.touch(exist_ok=True)



from typing import List, TYPE_CHECKING

from rich.console import Group
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TaskID
from yarl import URL

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


async def adjust_title(s: str, length: int = 40, placeholder: str = "...") -> str:
    """Collapse and truncate or pad the given string to fit in the given length"""
    return f"{s[:length - len(placeholder)]}{placeholder}" if len(s) >= length else s.ljust(length)


class ScrapingProgress:
    """Class that manages the download progress of individual files"""
    def __init__(self, visible_tasks_limit: int, manager: 'Manager'):
        self.manager = manager

        self.progress = Progress(SpinnerColumn(),
                                 "[progress.description]{task.description}")
        self.overflow = Progress("[progress.description]{task.description}")
        self.queue = Progress("[progress.description]{task.description}")
        self.progress_group = Group(self.progress, self.overflow, self.queue)

        self.color = "plum3"
        self.type_str = "Files"
        self.progress_str = "[{color}]{description}"
        self.overflow_str = "[{color}]... And {number} Other Links"
        self.queue_str = "[{color}]... And {number} Links In Scrape Queue"
        self.overflow_task_id = self.overflow.add_task(self.overflow_str.format(color=self.color, number=0, type_str=self.type_str), visible=False)
        self.queue_task_id = self.queue.add_task(self.queue_str.format(color=self.color, number=0, type_str=self.type_str), visible=False)

        self.visible_tasks: List[TaskID] = []
        self.invisible_tasks: List[TaskID] = []
        self.tasks_visibility_limit = visible_tasks_limit

    async def get_progress(self) -> Panel:
        """Returns the progress bar"""
        return Panel(self.progress_group, title="Scraping", border_style="green", padding=(1, 1))

    async def get_queue_length(self) -> int:
        """Returns the number of tasks in the scraper queue"""
        total = 0

        for scraper in self.manager.scrape_mapper.existing_crawlers.values():
            total += scraper.waiting_items

        return total

    async def redraw(self) -> None:
        """Redraws the progress bar"""
        while len(self.visible_tasks) > self.tasks_visibility_limit:
            task_id = self.visible_tasks.pop(0)
            self.invisible_tasks.append(task_id)
            self.progress.update(task_id, visible=False)
        while len(self.invisible_tasks) > 0 and len(self.visible_tasks) < self.tasks_visibility_limit:
            task_id = self.invisible_tasks.pop(0)
            self.visible_tasks.append(task_id)
            self.progress.update(task_id, visible=True)

        if len(self.invisible_tasks) > 0:
            self.overflow.update(self.overflow_task_id, description=self.overflow_str.format(color=self.color, number=len(self.invisible_tasks), type_str=self.type_str), visible=True)
        else:
            self.overflow.update(self.overflow_task_id, visible=False)

        queue_length = await self.get_queue_length()
        if queue_length > 0:
            self.queue.update(self.queue_task_id, description=self.queue_str.format(color=self.color, number=queue_length, type_str=self.type_str), visible=True)
        else:
            self.queue.update(self.queue_task_id, visible=False)

    async def add_task(self, url: URL) -> TaskID:
        """Adds a new task to the progress bar"""
        if len(self.visible_tasks) >= self.tasks_visibility_limit:
            task_id = self.progress.add_task(self.progress_str.format(color=self.color, description=str(url)), visible=False)
            self.invisible_tasks.append(task_id)
        else:
            task_id = self.progress.add_task(self.progress_str.format(color=self.color, description=str(url)))
            self.visible_tasks.append(task_id)
        await self.redraw()
        return task_id

    async def remove_task(self, task_id: TaskID) -> None:
        """Removes a task from the progress bar"""
        if task_id in self.visible_tasks:
            self.visible_tasks.remove(task_id)
            self.progress.update(task_id, visible=False)
        elif task_id in self.invisible_tasks:
            self.invisible_tasks.remove(task_id)
        elif task_id == self.overflow_task_id:
            self.overflow.update(task_id, visible=False)
        else:
            raise ValueError("Task ID not found")
        await self.redraw()
